/****c* sqlite3/sqlite3
 * NAME
 *    Slang Module for Sqlite version 3
 * CREATION DATE
 *    11/29/04 
 * AUTHOR
 *    Joe Robertson
 *    jmrobert5@mchsi.com
 * NOTES
 *    Originally started from template code.
 * HISTORY
 *    Second iteration of a module to interface the Sqlite 
 *    library.  This versionworks with the major revision
 *    to Sqlite, version 3.  Sqlite3 is much different from
 *    version 2.x.
 ***/

#include <stdio.h>
#include <slang.h>
#include "sqlite3.h"

SLANG_MODULE(sqlite3);

#define MODULE_MAJOR_VERSION	0
#define MODULE_MINOR_VERSION	0
#define MODULE_PATCH_LEVEL	2

static char *Module_Version_String = "0.0.2";

#define MODULE_VERSION_NUMBER	\
   (MODULE_MAJOR_VERSION*10000+MODULE_MINOR_VERSION*100+MODULE_PATCH_LEVEL)

/* Library interface functions */
typedef struct sqlite sqlite;
typedef struct sqlite_vm sqlite_vm;

/****d* sqlite/defines
 * DESCRIPTION
 * SQLITE_OK           0   /* Successful result 
 * SQLITE_ERROR        1   /* SQL error or missing database 
 * SQLITE_INTERNAL     2   /* An internal logic error in SQLite 
 * SQLITE_PERM         3   /* Access permission denied 
 * SQLITE_ABORT        4   /* Callback routine requested an abort 
 * SQLITE_BUSY         5   /* The database file is locked 
 * SQLITE_LOCKED       6   /* A table in the database is locked 
 * SQLITE_NOMEM        7   /* A malloc() failed 
 * SQLITE_READONLY     8   /* Attempt to write a readonly database 
 * SQLITE_INTERRUPT    9   /* Operation terminated by sqlite_interrupt() 
 * SQLITE_IOERR       10   /* Some kind of disk I/O error occurred 
 * SQLITE_CORRUPT     11   /* The database disk image is malformed 
 * SQLITE_NOTFOUND    12   /* (Internal Only) Table or record not found 
 * SQLITE_FULL        13   /* Insertion failed because database is full 
 * SQLITE_CANTOPEN    14   /* Unable to open the database file 
 * SQLITE_PROTOCOL    15   /* Database lock protocol error 
 * SQLITE_EMPTY       16   /* (Internal Only) Database table is empty 
 * SQLITE_SCHEMA      17   /* The database schema changed 
 * SQLITE_TOOBIG      18   /* Too much data for one row of a table 
 * SQLITE_CONSTRAINT  19   /* Abort due to contraint violation 
 * SQLITE_MISMATCH    20   /* Data type mismatch 
 * SQLITE_MISUSE      21   /* Library used incorrectly 
 * SQLITE_NOLFS       22   /* Uses OS features not supported on host 
 * SQLITE_AUTH        23   /* Authorization denied 
 * SQLITE_ROW         100  /* sqlite_step() has another row ready 
 * SQLITE_DONE        101  /* sqlite_step() has finished executing 
 *
 * sqlite3 slang module error codes 
 * SLANG_ARRAY_COPY_ERROR 901  /* module specific error code 
 *
 ***/

#define SQLITE_OK           0   /* Successful result */
#define SQLITE_ERROR        1   /* SQL error or missing database */
#define SQLITE_INTERNAL     2   /* An internal logic error in SQLite */
#define SQLITE_PERM         3   /* Access permission denied */
#define SQLITE_ABORT        4   /* Callback routine requested an abort */
#define SQLITE_BUSY         5   /* The database file is locked */
#define SQLITE_LOCKED       6   /* A table in the database is locked */
#define SQLITE_NOMEM        7   /* A malloc() failed */
#define SQLITE_READONLY     8   /* Attempt to write a readonly database */
#define SQLITE_INTERRUPT    9   /* Operation terminated by sqlite_interrupt() */
#define SQLITE_IOERR       10   /* Some kind of disk I/O error occurred */
#define SQLITE_CORRUPT     11   /* The database disk image is malformed */
#define SQLITE_NOTFOUND    12   /* (Internal Only) Table or record not found */
#define SQLITE_FULL        13   /* Insertion failed because database is full */
#define SQLITE_CANTOPEN    14   /* Unable to open the database file */
#define SQLITE_PROTOCOL    15   /* Database lock protocol error */
#define SQLITE_EMPTY       16   /* (Internal Only) Database table is empty */
#define SQLITE_SCHEMA      17   /* The database schema changed */
#define SQLITE_TOOBIG      18   /* Too much data for one row of a table */
#define SQLITE_CONSTRAINT  19   /* Abort due to contraint violation */
#define SQLITE_MISMATCH    20   /* Data type mismatch */
#define SQLITE_MISUSE      21   /* Library used incorrectly */
#define SQLITE_NOLFS       22   /* Uses OS features not supported on host */
#define SQLITE_AUTH        23   /* Authorization denied */
#define SQLITE_ROW         100  /* sqlite_step() has another row ready */
#define SQLITE_DONE        101  /* sqlite_step() has finished executing */

/* sqlite3 slang module error codes */
#define SLANG_ARRAY_COPY_ERROR 901  /* module specific error code */


/* global variables */
char *errmsg = 0; 
int nrows=0;
int ncols=0;
char **resultp = 0;
static sqlite3 *db = 0;
/*static sqlite_vm *vm = 0;*/
int ErrorCode = 0;

/* functions */
/****f* sqlite3/sqlite3_close
 * NAME
 *    sqlite3_close  -close the database.
 * ARGUMENTS
 *    none
 * RETURN VALUE
 *    int, error code
 * NOTES
 ***/
int close(){
    sqlite3_close(db);
}

/****f* sqlite3/sqlite3_errmsg
 * NAME
 *    sqlite3_errmsg  -get the last error message.
 * ARGUMENTS
 *    none
  * RETURN VALUE
 *    string, error message
 * NOTES
 ***/
char *get_errmsg(){
    return sqlite3_errmsg(db);
}

/****f* sqlite3/sqlite3_errcode
 * NAME
 *    sqlite3_errcode  -get the last error code.
 * ARGUMENTS
 *    none
  * RETURN VALUE
 *    int error code
 * NOTES
 ***/
int get_errcode(){
    return sqlite3_errcode(db);
}

/****f* sqlite3/sqlite3_exec
 * NAME
 *    sqlite3_exec  -execute a sql statement.
 * ARGUMENTS
 *    string, valid sqlite3 SQL string
 * OUTPUT
 *    changes to database
 * RETURN VALUE
 *    int, error code
 * SEE ALSO
 *    sqlite3_errcode, sqlite3_errmsg,   
 * NOTES
 ***/
int exec(const char *sql){
    int res = 0;

    /* db - we control
     * sql - input
     * callback - null
     * first arg to callback - null
     * errmsg - we control
     */
    res = sqlite3_exec(db, sql, NULL, NULL, &errmsg);
    return res;
}

/* this function is not exported */
/* this causes a problem, I need to double 
 * check if this is needed
 */
void free_table(){
    sqlite3_free_table(&errmsg);
}

/****f* sqlite3/sqlite3_changes
 * NAME
 *    sqlite3_changes  -get the number of changes caused
 *    by the last SQL action.
 * ARGUMENTS
 *    none
  * RETURN VALUE
 *    int, the number of changes in the db from the last SQL.
 * NOTES
 ***/
int changes(){
    return sqlite3_changes(db);
}

/****f* sqlite3/sqlite3_total_changes
 * NAME
 *    sqlite3_total_changes  -get the number of changes since the db was open.
 * ARGUMENTS
 *    none
  * RETURN VALUE
 *    int, the number of changes in the db.
 * NOTES
 ***/
int total_changes(){
    return sqlite3_total_changes(db);
}


/****f* sqlite3/sqlite3_get_rows
 * NAME
 *    sqlite3_get_rows  -get the table row count.
 * ARGUMENTS
 *    none
  * RETURN VALUE
 *    int, the number of rows in the table.
 * NOTES
 ***/
int get_rows(){ return nrows; }


/****f* sqlite3/sqlite3_get_cols
 * NAME
 *    sqlite3_get_cols  -get the table column count.
 * ARGUMENTS
 *    none
  * RETURN VALUE
 *    int, the number of columns in the table.
 * NOTES
 ***/
int get_cols(){ return ncols; }


/* this function is not exported. */
/* title row is always row 0 */
int get_cell(int row, int col){
    int index = 0;
    
    if ( row == 0) {
	index = col;
    }
    else {
	index = ((row *col)+col)-1;
    }
    return &resultp[index];
}


/****f* sqlite3/sqlite3_get_table
 * NAME
 *    sqlite3_get_table  -get a table of data results from a SQL query.
 * ARGUMENTS
 *    string, a sqlite3 SQL string statement.
 * RETURN VALUE
 *    an s-lang array is returned to the stack
 * EXAMPLE
 *    s-lang code
 *
 *    [open db]
 *    % query the table
 *    variable data;
 *    data = sqlite3_get_table("SELECT * FROM test;");
 *    vmessage("data size=%d", length(data) );
 *
 * SEE ALSO
 *    sqlite3_get_cols, sqlite3_get_rows
 * NOTES
 ***/
void get_table(const char *sql){
    int res = 0;
    int cnt = 0;
    SLang_Array_Type *at;
    int k = 0;

    res = sqlite3_get_table(db, sql, &resultp, &nrows, &ncols, &errmsg);

    if (res == SQLITE_OK){
	cnt = nrows * ncols;
	
	/* SLang_create_array (SLANG_STRING_TYPE, 0, NULL, &four, 1); 
	 * array (type, read-only(int), data(void *), *arr, #dimensions(int) ) 
	 */
	
	at = SLang_create_array(SLANG_STRING_TYPE, 0, NULL, &cnt, 1 );
	/* SLang_set_array_element(at, 0, "helloworld"); */
/* 	(void) SLang_push_array (at, 1); */

	if (at != NULL){
	    for (k=0; k<cnt; k++){
		SLang_set_array_element (at, &k, &resultp[k]);
	    }
	    
	    /* return the array to the stack */
	    /* also use the 1 arg to decrement refs by one */
	    (void) SLang_push_array (at, 1);
	}
	else {
	    /* set error code */
	    ErrorCode = SLANG_ARRAY_COPY_ERROR;
	    SLang_free_array (at);
	}
    }
    else {
	ErrorCode = res;
    }
    //free_table();
}


/****f* sqlite3/sqlite3_open
 * NAME
 *    sqlite3_open  -open a sqlite3 database.
 * ARGUMENTS
 *    string, a sqlite3 SQL string statement.
 * RETURN VALUE
 *    int, error code.
 * SEE ALSO
 *    sqlite3_errmsg
 * NOTES
 *    open will return 0 on success 
 *    non-0 on fail
 *    call sqlite3_errmsg if fail 
 ***/
int open(const char *dbname){
    return sqlite3_open(dbname, &db);
}


/* Define intrinsics here */
/* MAKE_INTRINSIC_input(slang-name, modulename, return val) */
static SLang_Intrin_Fun_Type Sqlite3_Intrinsics [] =
    {
	MAKE_INTRINSIC_0("sqlite3_close", close, SLANG_INT_TYPE),
	MAKE_INTRINSIC_0("sqlite3_changes", changes, SLANG_INT_TYPE),
	MAKE_INTRINSIC_0("sqlite3_errcode", get_errcode, SLANG_INT_TYPE),
	MAKE_INTRINSIC_0("sqlite3_errmsg", get_errmsg, SLANG_STRING_TYPE),
	MAKE_INTRINSIC_S("sqlite3_exec", exec, SLANG_INT_TYPE),
	MAKE_INTRINSIC_0("sqlite3_get_cols", get_cols, SLANG_INT_TYPE),
	MAKE_INTRINSIC_0("sqlite3_get_rows", get_rows, SLANG_INT_TYPE),
	MAKE_INTRINSIC_S("sqlite3_get_table", get_table, SLANG_VOID_TYPE),
	MAKE_INTRINSIC_S("sqlite3_open", open, SLANG_INT_TYPE),
	MAKE_INTRINSIC_0("sqlite3_total_changes", total_changes, SLANG_INT_TYPE),
	SLANG_END_INTRIN_FUN_TABLE
    };


static SLang_Intrin_Var_Type Sqlite3_Variables [] =
{
   MAKE_VARIABLE("_sqlite3_module_version_string", &Module_Version_String, SLANG_STRING_TYPE, 1),
   SLANG_END_INTRIN_VAR_TABLE
};


static SLang_IConstant_Type Sqlite3_Constants [] =
{
   MAKE_ICONSTANT("_sqlite3_module_version", MODULE_VERSION_NUMBER),
   SLANG_END_ICONST_TABLE
};


int init_sqlite3_module_ns (char *ns_name)
{
   SLang_NameSpace_Type *ns = SLns_create_namespace (ns_name);
   if (ns == NULL)
     return -1;

   if ((-1 == SLns_add_intrin_fun_table (ns, Sqlite3_Intrinsics, NULL))
       || (-1 == SLns_add_intrin_var_table (ns, Sqlite3_Variables, NULL))
       || (-1 == SLns_add_iconstant_table (ns, Sqlite3_Constants, NULL)))
     return -1;

   return 0;
}


/* This function is optional */
void deinit_sqlite3_module (void)
{
}
